import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';

enum SNSMobileSDKStatus {
  Ready,
  Initial,
  Incomplete,
  Pending,
  Approved,
  Failed,
  FinallyRejected,
  TemporarilyDeclined,
  ActionCompleted
}

enum SNSMobileSDKErrorType {
  Unknown,
  InvalidParameters,
  Unauthorized,
  InitialLoadingFailed,
  ApplicantNotFound,
  ApplicantMisconfigured,
  NetworkError,
  UnexpectedError,
  InititlizationError,
}

enum SNSFlowActionType {

  Unknown,
  FaceEnrollment
}

enum SNSMobileSDKAnswerType {

  Ignored,
  Red,
  Yellow,
  Green,
  Error,
}

enum SNSActionResultHandlerReaction {

  Continue,
  Cancel,
}

typedef SNSTokenExpirationHandler = Future<String> Function();
typedef SNSReady = Function();
typedef SNSStatusChangedHandler = Function(SNSMobileSDKStatus newStatus, SNSMobileSDKStatus oldStatus);
typedef SNSActionResultHandler = Future<SNSActionResultHandlerReaction> Function(SNSMobileSDKActionResult result);
typedef SNSErrorHandler = Function(SNSException exception);

class SNSMobileSDK {

  MethodChannel _channel = MethodChannel('sumsub.com/flutter_idensic_mobile_sdk_plugin');

  static SNSMobileSDKBuilder builder(String apiUrl, String flowName) => SNSMobileSDKBuilder(apiUrl: apiUrl, flowName: flowName);

  final String apiUrl;
  final String flowName;
  final String accessToken;
  final Locale locale;
  final String supportEmail;
  final SNSTokenExpirationHandler onTokenExpiration;
  final SNSStatusChangedHandler onStatusChanged;
  final SNSActionResultHandler onActionResult;
  final bool isDebug;

  SNSMobileSDK._builder(SNSMobileSDKBuilder builder) :
    apiUrl = builder.apiUrl,
    flowName = builder.flowName,
    accessToken = builder.accessToken,
    locale = builder.locale,
    supportEmail = builder.supportEmail,
    onTokenExpiration = builder.onTokenExpiration,
    onStatusChanged = builder.onStatusChanged,
    onActionResult = builder.onActionResult,
    isDebug = builder.isDebug;

  void dismiss() {
    _channel.invokeMethod('dismiss');
  }

  Future<SNSMobileSDKResult> launch() async {
    _channel.setMethodCallHandler(_onMethodCallHandler);
    final Map<dynamic, dynamic> map = await _channel.invokeMethod('onLaunchSDK', _toArguments());
    return Future.value(SNSMobileSDKResult.fromMap(map));
  }

  Future<String> _onMethodCallHandler(MethodCall call) async {
    switch(call.method) {
      case "onTokenExpiration":
        return onTokenExpiration();
      case "onError":
        return null;
      case "onStatusChanged":
        return onStatusChanged(_toStatus(call.arguments[0]), _toStatus(call.arguments[1]));
      case "onActionResult":
        final SNSActionResultHandlerReaction reaction = await onActionResult(_toActionResult(call.arguments[0]));
        return Future.value(_fromActionResultHandlerReaction(reaction));
      default:
        print('Unknown method ${call.method}');
        throw MissingPluginException();
    }
  }

  Map<String, dynamic> _toArguments() {
    return {
      "apiUrl": apiUrl,
      "flowName": flowName,
      "accessToken": accessToken,
      "languageCode": locale?.languageCode,
      "supportEmail": supportEmail,
      "isDebug": isDebug,
      "hasOnStatusChanged": onStatusChanged != null,
      "hasOnActionResult": onActionResult != null,
    };
  }
}

class SNSMobileSDKBuilder {
  final String apiUrl;
  final String flowName;

  // builder
  String accessToken;
  Locale locale;
  String supportEmail;
  SNSTokenExpirationHandler onTokenExpiration;
  SNSStatusChangedHandler onStatusChanged;
  SNSActionResultHandler onActionResult;
  bool isDebug;

  SNSMobileSDKBuilder({@required this.apiUrl, @required this.flowName});

  SNSMobileSDKBuilder withAccessToken(String accessToken, SNSTokenExpirationHandler onTokenExpiration) {
    this.accessToken = accessToken;
    this.onTokenExpiration = onTokenExpiration;
    return this;
  }

  SNSMobileSDKBuilder withLocale(Locale locale) {
    this.locale = locale;
    return this;
  }

  SNSMobileSDKBuilder withSupportEmail(String supportEmail) {
    this.supportEmail = supportEmail;
    return this;
  }

  SNSMobileSDKBuilder withHandlers({SNSStatusChangedHandler onStatusChanged, SNSActionResultHandler onActionResult}) {
    this.onStatusChanged = onStatusChanged;
    this.onActionResult = onActionResult;
    return this;
  }

  SNSMobileSDKBuilder withDebug(bool isDebug) {
    this.isDebug = isDebug;
    return this;
  }

  SNSMobileSDK build() => SNSMobileSDK._builder(this);
}

class SNSMobileSDKResult {

  final bool success;
  final SNSMobileSDKStatus status;
  final SNSMobileSDKErrorType errorType;
  final String errorMsg;
  final SNSMobileSDKActionResult actionResult;

  SNSMobileSDKResult(this.success, this.status, this.errorType, this.errorMsg, this.actionResult);

  static SNSMobileSDKResult fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSMobileSDKResult(map['success'] as bool, _toStatus(map['status']), _toErrorType(map['errorType']), map['errorMsg'], _toActionResult(map['actionResult']));
  }

  String toString() => 'SNSMobileSDKResult(success: $success, status: $status, errorType: $errorType, errorMsg: $errorMsg, actionResult: $actionResult)';

  @override
  bool operator ==(Object o) {
    if (identical(this, o)) return true;
  
    return o is SNSMobileSDKResult &&
      o.success == success &&
      o.status == status &&
      o.errorType == errorType &&
      o.errorMsg == errorMsg &&
      o.actionResult == actionResult;
  }

  @override
  int get hashCode {
    return success.hashCode ^
      status.hashCode ^
      errorType.hashCode ^
      errorMsg.hashCode ^
      actionResult.hashCode;
  }
}

class SNSMobileSDKActionResult {

  final String actionId;
  final String actionType;
  final SNSMobileSDKAnswerType answer;
  final bool allowContinuing;

  SNSMobileSDKActionResult(this.actionId, this.actionType, this.answer, this.allowContinuing);

  @override
  String toString() => 'SNSMobileSDKActionResult(actionId: $actionId, answer: $answer, actionType: $actionType, allowContinuing: $allowContinuing)';

  @override
  bool operator ==(Object o) {
    if (identical(this, o)) return true;
  
    return o is SNSMobileSDKActionResult &&
      o.actionId == actionId &&
      o.answer == answer;
  }

  @override
  int get hashCode => actionId.hashCode ^ answer.hashCode;
}

class SNSException {

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    switch (map['type']) {
      case "Api":
        return SNSApiException.fromMap(map['payload']);
      case "Network":
        return SNSNetworkException.fromMap(map['payload']);
      case "Unknown":
        return SNSUnknownException.fromMap(map['payload']);
    }

    throw new Exception("Unknown exception type (${map['type']}");
  }
}

class SNSApiException extends SNSException {

  final int code;
  final String description;
  final String correlationId;

  SNSApiException(this.code, this.description, this.correlationId);

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSApiException(map['code'] as int, map['description'] as String, map['correlationId'] as String);
  }  

  @override
  String toString() => 'SNSApiException(code: $code, description: $description, correlationId: $correlationId)';
}

class SNSNetworkException extends SNSException {

  final String message;

  SNSNetworkException(this.message);

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSNetworkException(map['message'] as String);
  }

  @override
  String toString() => 'SNSNetworkException(message: $message)';
}

class SNSUnknownException extends SNSException {

  final String message;

  SNSUnknownException(this.message);

  static SNSException fromMap(Map<dynamic, dynamic> map) {
    if (map == null) return null;

    return SNSUnknownException(map['message'] as String);
  }

  @override
  String toString() => 'SNSUnknownException(message: $message)';
}

SNSMobileSDKStatus _toStatus(String value) {
  SNSMobileSDKStatus status;
  switch (value) {
    case 'Ready': 
      status = SNSMobileSDKStatus.Ready;
      break;
    case 'Initial': 
      status = SNSMobileSDKStatus.Initial;
      break;
    case 'Incomplete': 
      status = SNSMobileSDKStatus.Incomplete;
      break;
    case 'Pending': 
      status = SNSMobileSDKStatus.Pending;
      break;
    case 'Approved': 
      status = SNSMobileSDKStatus.Approved;
      break;
    case 'Failed': 
      status = SNSMobileSDKStatus.Failed;
      break;
    case 'FinallyRejected': 
      status = SNSMobileSDKStatus.FinallyRejected;
      break;
    case 'TemporarilyDeclined': 
      status = SNSMobileSDKStatus.TemporarilyDeclined;
      break;
    case 'ActionCompleted':
      status = SNSMobileSDKStatus.ActionCompleted;
      break;
    default:
      break;
  }

  return status;
}

SNSMobileSDKErrorType _toErrorType(String value) {
  SNSMobileSDKErrorType errorType;
  switch (value) {
    case 'Unknown': 
      errorType = SNSMobileSDKErrorType.Unknown;
      break;
    case 'InvalidParamaters': 
      errorType = SNSMobileSDKErrorType.InvalidParameters;
      break;
    case 'Unauthorized': 
      errorType = SNSMobileSDKErrorType.Unauthorized;
      break;
    case 'InitialLoadingFailed': 
      errorType = SNSMobileSDKErrorType.InitialLoadingFailed;
      break;
    case 'ApplicantNotFound': 
      errorType = SNSMobileSDKErrorType.ApplicantNotFound;
      break;
    case 'ApplicantMisconfigured': 
      errorType = SNSMobileSDKErrorType.ApplicantMisconfigured;
      break;
    case 'NetworkError': 
      errorType = SNSMobileSDKErrorType.NetworkError;
      break;
    case 'UnexpectedError': 
      errorType = SNSMobileSDKErrorType.UnexpectedError;
      break;
    case 'InititlizationError': 
      errorType = SNSMobileSDKErrorType.InititlizationError;
      break;      
    default:
      break;
  }

  return errorType;
}

// ignore: unused_element
SNSFlowActionType _toFlowActionType(String value) {
  SNSFlowActionType type;
  switch (value) {
    case 'faceEnrollment': 
      type = SNSFlowActionType.FaceEnrollment;
      break;
    case 'none': 
    default:
      type = SNSFlowActionType.Unknown;
      break;
  }

  return type;
}

SNSMobileSDKAnswerType _toAnswerType(String value) {
  SNSMobileSDKAnswerType type;
  switch (value) {
    case 'IGNORED':
      type = SNSMobileSDKAnswerType.Ignored;
      break;
    case 'RED':
      type = SNSMobileSDKAnswerType.Red;
      break;
    case 'YELLOW':
      type = SNSMobileSDKAnswerType.Yellow;
      break;
    case 'GREEN':
      type = SNSMobileSDKAnswerType.Green;
      break;
    case 'ERROR':
      type = SNSMobileSDKAnswerType.Error;
      break;
  }

  return type;
}

SNSMobileSDKActionResult _toActionResult(Map<dynamic, dynamic> value) {
  if (value == null) {
    return null;
  }

  return SNSMobileSDKActionResult(value['actionId'], value['actionType'], _toAnswerType(value['answer']), value['allowContinuing']);
}

String _fromActionResultHandlerReaction(SNSActionResultHandlerReaction reaction) {
  switch (reaction) {
    case SNSActionResultHandlerReaction.Cancel:
      return "cancel";
    default:
      return "continue";
  }
}